/*
 * Copyright (c) 1997 Shigio Yamaguchi. All rights reserved.
 * Copyright (c) 1999 Tama Communications Corporation. All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */
#ifdef	HAVE_CONFIG_H
#include <config.h>
#endif/*HAVE_CONFIG_H_*/

#include <os.h>
#include <errno.h>
#include <stdlib.h>
#include <ucs2char.h>

/*
 * rel2abs: convert an relative path name into absolute.
 *
 *	i)	path	relative path
 *	i)	base	base directory (must be absolute path)
 *	o)	result	result buffer
 *	i)	size	size of result buffer
 *	r)		!= NULL: absolute path
 *			== NULL: error
 */
ucs2_char_t *
rel2abs(path, base, result, size)
	const ucs2_char_t *path;
	const ucs2_char_t *base;
	ucs2_char_t *result;
	const size_t size;
{
	static const ucs2_char_t ucs2cs_dot[] = {'.',0};
	static const ucs2_char_t ucs2cs_dot_dot[] = {'.','.',0};
	static const ucs2_char_t ucs2cs_dot_slash[] = {'.','/',0};
	static const ucs2_char_t ucs2cs_dot_dot_slash[] = {'.','.','/',0};
	const ucs2_char_t *pp, *bp;
	/*
	 * endp points the last position which is safe in the result buffer.
	 */
	const ucs2_char_t *endp = result + size - 1;
	ucs2_char_t *rp;
	int length;

	if (*path == '/') {
		if (ucs2len(path) >= size)
			goto erange;
		ucs2cpy(result, path);
		goto finish;
	} else if (*base != '/' || !size) {
		errno = EINVAL;
		return (NULL);
	} else if (size == 1)
		goto erange;

	length = ucs2len(base);

	if (!ucs2cmp(path, ucs2cs_dot) || !ucs2cmp(path, ucs2cs_dot_slash)) {
		if (length >= size)
			goto erange;
		ucs2cpy(result, base);
		/*
		 * rp points the last char.
		 */
		rp = result + length - 1;
		/*
		 * remove the last '/'.
		 */
		if (*rp == '/') {
			if (length > 1)
				*rp = 0;
		} else
			rp++;
		/* rp point NULL char */
		if (*++path == '/') {
			/*
			 * Append '/' to the tail of path name.
			 */
			*rp++ = '/';
			if (rp > endp)
				goto erange;
			*rp = 0;
		}
		goto finish;
	}
	bp = base + length;
	if (*(bp - 1) == '/')
		--bp;
	/*
	 * up to root.
	 */
	for (pp = path; *pp && *pp == '.'; ) {
		if (!ucs2ncmp(pp, ucs2cs_dot_dot_slash, 3)) {
			pp += 3;
			while (bp > base && *--bp != '/')
				;
		} else if (!ucs2ncmp(pp, ucs2cs_dot_slash, 2)) {
			pp += 2;
		} else if (!ucs2ncmp(pp, ucs2cs_dot_dot, 3)) {
			pp += 2;
			while (bp > base && *--bp != '/')
				;
		} else
			break;
	}
	/*
	 * down to leaf.
	 */
	length = bp - base;
	if (length >= size)
		goto erange;
	ucs2ncpy(result, base, length);
	rp = result + length;
	if (*pp || *(pp - 1) == '/' || length == 0)
		*rp++ = '/';
	if (rp + ucs2len(pp) > endp)
		goto erange;
	ucs2cpy(rp, pp);
finish:
	return result;
erange:
	errno = ERANGE;
	return (NULL);
}
